:- module(kategorien,[kategorie/1]).

% ----- Wertebereiche der Art- und Formmerkmale -----

genus(Gen) :-                         % Nomen u.a.
    (Gen = fem ; Gen = mask ; Gen = neut).
numerus(X) :-
    (X = sg ; X = pl).
kasus(X) :-
    (X = nom ; X = gen ; X = dat ; X = akk).

person(X) :- (X = 1 ; X = 2 ; X = 3). % Pronomen, Verb
tempus(X) :- (X = praes ; X = praet). % Verb
modus(X) :- (X = ind ; X = konj).     % Verb

% ----- Kategorie(Artmerkmale,Formmerkmale) -----

kategorie(n([Gen],[Num,Kas])) :-
    genus(Gen), numerus(Num), kasus(Kas).
kategorie(rn([Gen],[Num,Kas])) :-
    genus(Gen), numerus(Num), kasus(Kas).

kategorie(en([Gen],[sg,Kas])) :-
    genus(Gen), kasus(Kas).

kategorie(pron([Def],[Gen,Num,Kas])) :- % nur 3.Person!
    (Def = per ; Def = rel ; Def = qu), % er, der, wer
    genus(Gen), numerus(Num), kasus(Kas).
    
kategorie(det([Def],[Gen,Num,Kas])) :-
    % ein/der,welcher, jeder/alle,viele,die meisten
    member(Def,[indef,def,qu,all,many,most]), 
    genus(Gen), numerus(Num), kasus(Kas).

kategorie(poss([def(G,N)],[Gen,Num,Kas])) :-
    genus(G), numerus(N), % sein/ihr
    genus(Gen), numerus(Num), kasus(Kas).
kategorie(poss([rel(G,N)],[-,-,-])) :-
    genus(G), numerus(N). % dessen/deren
kategorie(poss([qu],[-,-,-])). % wessen

kategorie(np([Def,Pers,Gen,Num],[Kas])) :-  
    definitheit(np,Def), % Num inhaerent
    Pers = 3, % TODO für Dialoge: 1.+2.Person
    genus(Gen), numerus(Num), kasus(Kas).

definitheit(np,Def) :- 
    ( Def = def ; Def = indef ; Def = quant 
    ; Def = qu ; Def = refl ; Def = rezi       
    ; Def = rel(Gen,Num), genus(Gen), numerus(Num) ).

kategorie(v([nom],[Pers,Num,Temp,Mod])) :-
    Pers = 3,      % TODO: 1.2 Person fuer Dialoge
    numerus(Num), tempus(Temp), modus(Mod). % Mod := ind

kategorie(v([nom,akk],[Pers,Num,Temp,Mod])) :-
    Pers = 3,      % TODO: 1.2 Person fuer Dialoge
    numerus(Num), tempus(Temp), modus(Mod). % Mod := ind

kategorie(v([H],[Pers,Num,Temp,Mod])) :-
    hilfsverb(H),
    Pers = 3, numerus(Num), tempus(Temp), modus(Mod).
hilfsverb(H) :- (H = sein ; H = werden). 

kategorie(s([Def],[Temp,Mod,Vst])) :- % einfache Saetze
    definitheit(s,Def),
    tempus(Temp), % ggf. ; Temp = perf
    modus(Mod),
    verbstellung(Vst).

definitheit(s,Def) :-
    ( Def = def ; Def = qu
    ; Def = rel(Gen,Num), genus(Gen),numerus(Num) ).
verbstellung(Vst) :-
    (Vst = ve ; Vst = vz ; Vst = vl).

