/* Zweck: Stelle Baumterme mit dem Programm /bin/dot graphisch dar.
   Autor: H.Leiss,                    29.5.2016  fuer Baueme mit Semantik

   Siehe das Handbuch dazu unter Unix: man dot  ...

   Man bekommt dot von AT&T als Teil des Graphikpakets graphviz unter
      http://www.research.att.com/sw/tools/graphviz/download.html

   Ein Baumterm ist eine Liste [Cat|SubTrees], worin
   - Cat eine einfache Kategorie der Grammatik ist
     (Kategoriename als Funktionszeichen, Merkmale als Argumente),
   - SubTrees die Baumterme der direkten Teilbäume sind.

   Der Baum ordnet die Blätter auf einer Grundlinie entsprechend den Wörtern
   des Eingabeausdrucks an.

*/

:- module(display_semtree,[displaySemTrees/2,displaySemTree/2,
                           displaySemTyTrees/3,displaySemTyTree/3]).
                        
/* Schreibe eine Liste <Terme> von Prolog-Termen als Bäume im .dot-Format auf die
   Datei <Datei>.tmp.dot, mit Sem/2 zur Berechnung der Bedeutungen : */

showSemTrees(Terme,Sem/2,Datei) :-
        concat(Datei,'.tmp.dot',File),
        tell(File),
        dotSem(Terme,Sem/2),
        told.

dotSem([Term|Terme],Sem/2) :-
        nl,
        (ground(Term) ->
            write('digraph syntaxbaum {'),
            write('\ncenter=true; size=\"7.5,14\";\c
                   \nratio=compress; ordering=out;\c
                   \nedge[dir=none]; node[shape=plaintext];\n'),
            drawSemTree(Term,Sem/2),
            write('}')
        ; write(user,'Fehler: der darzustellende Term \"'),
            writeq(user,Term),
            write(user, '\" enthält freie Variable!\n')
        ),
        dotSem(Terme,Sem/2).
dotSem([],_).

drawSemTree(Term,Sem/2) :-
    drawSemTree(Term,Sem/2,0,_,[],[],BlattfolgeR,[]),
    reverse(BlattfolgeR,Blattfolge),
    order(Blattfolge).

order(Blattfolge) :-
    write('{ rank=same ; '),
    write_list(Blattfolge,' ; '),
    write('};\n'),
    write('subgraph {edge[style=invis]; '), % Blaetter
    write_list(Blattfolge,' -> '),          % auf einer Linie
    write(' }\n').                          % anordnen

write_list([A,B|Bs],Atom) :-
    write(A),write(Atom),write_list([B|Bs],Atom).
write_list([A],_) :-
    write(A).
write_list([],_).

/* drawTree(+Term,+Sem/2,+Wurzelnummer,-naechste unbenutzte Nummer,4 Akkumulatoren)  */

drawSemTree([Term],_/2,N,K,AkkL,AkkR,[N|AkkL],AkkR) :-
    atom(Term),
    !,
    write(N),
    write(' [fontcolor = blue, label = "'),write(Term),write('"];'),
    write('\n'),
    K is N+1.

drawSemTree([Kategorie,(Arg1,Arg2)],Sem/2,N,K,AkkL,_,NAkkL,NAkkR) :-
    !, % only for trees with discontinuous constituents?
    write(N),
    write(' [label = "'),write(Kategorie),write('"];\n'),
    % Semantikterme ergaenzen
    M is N+1,
    drawSemTrees([Arg1],Sem/2,N,M,L,AkkL,[],NAkkL,[]),
    drawSemTrees([Arg2],Sem/2,N,L,K,[],[],NAkkR,[]).

drawSemTree([Kategorie|Args],Sem/2,N,K,AkkL,AkkR,AkkLN,AkkRN) :-
    !,
    write(N),write(' [label = \"'),write(Kategorie),
    % alle Lambda-Terme unter die Kategorie schreiben!!!  % semTy, semDRS ebenso?
    Eval =.. [Sem,[Kategorie|Args],LamTerm],
    ( call(Eval), % sem([Kategorie|Args],LamTerm),
        write('\\n + '), writen(LamTerm), fail ; true ),
    write('\"];\n'),
    M is N+1,
    drawSemTrees(Args,Sem/2,N,M,K,AkkL,AkkR,AkkLN,AkkRN).

list([]).
list([_|L]) :- list(L).

/* Nachfolger von Baum N, ab Nummer M bis Nummer K fuer deren Knoten */

drawSemTrees([Arg|Args],Sem/2,N,M,K,AkkL,AkkR,NAkkL,NAkkR) :-
    write(N), write(' -> '), write(M), write(' ;\n'),        
    drawSemTree(Arg,Sem/2,M,K1,AkkL,AkkR,NAkkL1,NAkkR1),
    drawSemTrees(Args,Sem/2,N,K1,K,NAkkL1,NAkkR1,NAkkL,NAkkR).

drawSemTrees([],_,_,M,M,AkkL,AkkR,AkkL,AkkR).

/* Anzeigen der Bäume am Bildschirm: */

displaySemTree(Term,Sem/2) :-
    displaySemTrees([Term],Sem/2).

displaySemTrees(Terme,Sem/2) :-
    showSemTrees(Terme,Sem/2,syntaxbaum),
    viewgraph(syntaxbaum).

viewgraph(Datei) :-
    concat_atom(['dot -Tps ',Datei,'.tmp.dot ',' -o',Datei,'.tmp.ps '],Command1),
    % Kleiner: concat_atom(['gv -magstep -2 ',Datei,'.tmp.ps &\n'],Command2),
    concat_atom(['gv -magstep=2 ',Datei,'.tmp.ps &\n'],Command2),
    shell(Command1),
    shell(Command2).

/* Beispiel: Baumkodierung wie in CL-2, nicht wie in Satzgrammatik

displayTree([np([def, 3, mask], [sg, nom]),
               [det([def], [mask, sg, nom]), [der]],
               [rn([mask], [sg, nom]), ['Planet']],
               [np([def, 3, fem], [sg, gen]),
                  [det([def], [fem, sg, gen]), [der]],
                  [en([fem], [sg, nom]), ['Venus']]]]).

*/


/* Schreibe eine Liste <Terme> von Prolog-Termen als Bäume im .dot-Format auf die
   Datei <Datei>.tmp.dot, mit Sem/2,SemTy/2 zur Berechnung getypter Bedeutungen : */

showSemTrees(Terme,Sem/2,SemTy/2,Datei) :-
        concat(Datei,'.tmp.dot',File),
        tell(File),
        dotSem(Terme,Sem/2,SemTy/2),
        told.

dotSem([Term|Terme],Sem/2,SemTy/2) :-
        nl,
        (ground(Term) ->
            write('digraph syntaxbaum {'),
            write('\ncenter=true; size=\"7.5,14\";\c
                   \nratio=compress; ordering=out;\c
                   \nedge[dir=none]; node[shape=plaintext];\n'),
            drawSemTree(Term,Sem/2,SemTy/2),
            write('}')
        ; write(user,'Fehler: der darzustellende Term \"'),
            writeq(user,Term),
            write(user, '\" enthält freie Variable!\n')
        ),
        dotSem(Terme,Sem/2,SemTy/2).
dotSem([],_Sem/2,_SemTy/2).

drawSemTree(Term,Sem/2,SemTy/2) :-
    drawSemTree(Term,Sem/2,SemTy/2,0,_,[],[],BlattfolgeR,[]),
    reverse(BlattfolgeR,Blattfolge),
    order(Blattfolge).

/* drawTree(+Term,+Sem/2,+SemTy/2,+Wurzelnummer,-naechste unbenutzte Nummer,4 Akkumulatoren)  */

drawSemTree([Term],_/2,_/2,N,K,AkkL,AkkR,[N|AkkL],AkkR) :-
    atom(Term),
    !,
    write(N),
    write(' [fontcolor = blue, label = "'),write(Term),write('"];'),
    write('\n'),
    K is N+1.

% drawSemTree([Kategorie,(Arg1,Arg2)],Sem/2,SemTy/2,N,K,AkkL,_,NAkkL,NAkkR) :-
%     !, % only for trees with discontinuous constituents?
%     write(N),
%     write(' [label = "'),write(Kategorie),write('"];\n'),
%     % Semantikterme ergaenzen
%     M is N+1,
%     drawSemTrees([Arg1],Sem/2,SemTy/2,N,M,L,AkkL,[],NAkkL,[]),
%     drawSemTrees([Arg2],Sem/2,SemTy/2,N,L,K,[],[],NAkkR,[]).

drawSemTree([Kategorie|Args],Sem/2,SemTy/2,N,K,AkkL,AkkR,AkkLN,AkkRN) :-
    !,
    write(N),write(' [label = \"'),write(Kategorie),
    % alle Lambda-Terme unter die Kategorie schreiben!!!  % semTy, semDRS ebenso?
    Eval =.. [Sem,[Kategorie|Args],LamTerm],
    SemTyExists =.. [SemTy,[Kategorie|Args],_],
    EvalTy =.. [SemTy,[Kategorie|Args],LamTermTy],
    ( SemTyExists
    -> ( call(EvalTy), 
           write('\\n + '), writen(LamTermTy), fail ; true )
    ;  ( call(Eval),
           write('\\n - '), writen(LamTerm), fail ; true )
    ),
    write('\"'), ( SemTyExists -> true ; write(', fontcolor=brown') ),
    write('];\n'),
    M is N+1,
    (debugging(resolve) % reverse the sentence sequence of a discourse
    -> swapSentences([Kategorie|Args],[Kategorie|ArgsSw])
    ; ArgsSw = Args),
    drawSemTrees(ArgsSw,Sem/2,SemTy/2,N,M,K,AkkL,AkkR,AkkLN,AkkRN).

swapSentences([d([def])],[d([def])]) :- !.
swapSentences([d([def]),S,D],[d([def]),DRev,S]) :-
   S = [s(_,_)|_], D = [d(_)|_], 
   !, swapSentences(D,DRev).
swapSentences([Kat|Args],[Kat|Args]).

/* Nachfolger von Baum N, ab Nummer M bis Nummer K fuer deren Knoten */

drawSemTrees([Arg|Args],Sem/2,SemTy/2,N,M,K,AkkL,AkkR,NAkkL,NAkkR) :-
    write(N), write(' -> '), write(M), write(' ;\n'),        
    drawSemTree(Arg,Sem/2,SemTy/2,M,K1,AkkL,AkkR,NAkkL1,NAkkR1),
    drawSemTrees(Args,Sem/2,SemTy/2,N,K1,K,NAkkL1,NAkkR1,NAkkL,NAkkR).

drawSemTrees([],_,_,_,M,M,AkkL,AkkR,AkkL,AkkR).

/* Anzeigen der Bäume am Bildschirm: */

displaySemTyTree(Term,Sem/2,SemTy/2) :-
    displaySemTyTrees([Term],Sem/2,SemTy/2).

displaySemTyTrees(Terme,Sem/2,SemTy/2) :-
    showSemTrees(Terme,Sem/2,SemTy/2,syntaxbaum),
    viewgraph(syntaxbaum).
