:- module(tokenizer,[read_sentence/3,tokenize/2,
                     satzende/0,parse/0,parsed/0,parse/1,is_comment/1]).
% 22.9.2015: Neue Satzenden ":\n" und "!\n", is_comment/1

% read_sentence(+Stream,-Sentence,+Seen)
read_sentence(Stream,Sentence,Seen) :-
    get0(Stream,Char),  
    read_sentence(Stream,Sentence,Char,Seen).

read_sentence(Stream,Sentence,Char,Seen) :-
    (Char = -1  % Dateiende
    -> reverse(Seen,Sentence)
    ; (Seen = [C|Cs], satzende([C,Char]))
    -> reverse(Cs,Sentence)
    ; Char = 10 % return/newline ignorieren
    -> read_sentence(Stream,Sentence,Seen)
    ; read_sentence(Stream,Sentence,[Char|Seen])).

% read_discourse(+Stream,+Seen,+Saetze,-Discourse)
read_discourse(Stream,Seen,Saetze,Discourse) :-
    get0(Stream,Char), 
    read_discourse(Stream,Char,Seen,Saetze,Discourse).

read_discourse(Stream,Char,Seen,Saetze,Discourse) :-
    (Char = -1  % Dateiende
    -> (reverse(Seen,Satz),
        (Saetze = [] -> append(Satz,Discourse)
        ; append(Satz,[32|Saetze],Discourse)))
    ; (Seen = [C|Cs], satzstop([C,Char]))
    -> (reverse(Cs,Satz), 
        append([32|Satz],Saetze,Discourse1),
	read_discourse(Stream,[],Discourse1,Discourse))
    ; (Seen = [C|Cs], satzende([C,Char]))
    -> (reverse(Cs,Satz) , append(Satz,Saetze,Discourse))
    ; Char = 10 % return/newline ignorieren
    -> read_discourse(Stream,Seen,Saetze,Discourse)
    ; read_discourse(Stream,[Char|Seen],Saetze,Discourse)).

satzstop([46,32]).  % <Punkt><Space>
satzstop([63,32]).  % <Fragezeichen><Space>
satzende([46,10]).  % <Punkt><Return>
satzende([58,10]).  % <Doppelpunkt><Return>
satzende([38,10]).  % <Ausrufezeichen><Return>

% tokenize(+String,-Atomlist).
tokenize(String,Atomlist) :-
    name(Atom,String),
    concat_atom(AtomlistK,' ',Atom),
    % = atomic_list_concat(AtomlistK,' ',Atom)
    entferne_komma(AtomlistK,Atomlist).

entferne_komma([AtomK|AtomeK],[Atom|Atome]) :-
    (( atom_concat(Atom,',',AtomK)
     ; atom_concat(Atom,'.',AtomK) % Diskurs
     )                          
    -> true
    ; Atom = AtomK),
    entferne_komma(AtomeK,Atome).
entferne_komma([],[]).
is_comment([First|_]) :- First =  37. % '%'

satzende :-
    write('Beende die Eingabe mit <Punkt><Return>'),
    nl,read_sentence(user,Sentence,[]),
    name(Satz,Sentence),
    nl,write('Eingabe:   '),writeq(Satz),
    tokenize(Sentence,Atomlist),
    nl,write('Atomliste: '),writeq(user,Atomlist).

parse :-
    write('Beende die Eingabe mit <Punkt><Return>'),
    nl,read_sentence(user,Sentence,[]),
    tokenize(Sentence,Atomlist),
    startsymbol(S),
    addTree:translate(S,Term,[Baum]),  
    addDifflists:translate(Term,TermExp,Atomlist,[]),
    nl,write('Aufruf: '), portray_clause(TermExp),
    call(TermExp), 
    write('Baum:   '), nl,showTree(Baum,8),nl,
    fail.
parse.
parsed :-
    write('Beende die Eingabe mit <Punkt><Return>'),
    nl,read_sentence(user,Sentence,[]),
    tokenize(Sentence,Atomlist),
    setof(Baum, parse(Atomlist, Baum), Trees),
    displayTrees(Trees).

parse(Atomlist,Baum) :-
    startsymbol(S),
    addTree:translate(S,Term,[Baum]),
    addDifflists:translate(Term,Exp,Atomlist,[]),
    call(Exp).
    
parse(Dateiname) :-
   % atom_concat(Dateiname,'.aus',Ausgabe),
   % tell(Ausgabe), % ggf. Ausgabedatei öffnen
   open(Dateiname,read,Strom),
   parsen(Strom),
   close(Strom),
   % told,          % ggf. Ausgabedatei schließen
   nl,write('Fertig.').

parsen(Strom) :-
   read_sentence(Strom, Satz, []),
   (Satz = [] -> true  % Dateiende 
   ; is_comment(Satz) ->
       atom_codes(Atom,Satz),nl,write(Atom),
       parsen(Strom)
   ; tokenize(Satz,Atome),
       nl,writeq(Atome),nl,
       (setof(Baum,parse(Atome,Baum),Trees)
       -> showTrees(Trees,3), nl
       ; nl, tab(3),
           write('* keine Analysen gefunden *'),
           nl),
       parsen(Strom)   % weitere Sätze
   ).

