% mehrere Terme: evals(+Belegung,+Terme,-Objekte)
evals(Env,[Term|Terms],[Val|Vals]) :-
    eval(Env,Term,Val), evals(Env,Terms,Vals).
evals(_Env,[],[]).
    
% eval(+Belegung,+Term,-Objekt), (Rekursion bzgl.Term)
eval([(Var,Obj)|Env],X,Val) :-
    var(X),
    !, (Var == X -> Val = Obj ; eval(Env,X,Val)).
eval([],X,_) :-
    var(X),
    !, format('\nFehler: unbelegte Variable').
eval(_Env,C,Val) :-
    functor(C,C,0), !, Val = C.

eval(Env,card(X:Ty,Restr,Formel),N) :-
    var(X),!,
    findall(O,(objekt(O:Ty),
               eval([(X,O)|Env],(Restr & Formel))),Os),
    sort(Os,Objs), length(Objs,N).
eval(Env,card((X,Y):TyX*TyY,Restr,Formel),N) :-
    var(X),var(Y),!,
    findall((A,B),(objekt(A:TyX),objekt(B:TyY),
               eval([(X,A),(Y,B)|Env],(Restr & Formel))),
            ABs), sort(ABs,Pairs), length(Pairs,N).

% eval(+Belegung,+Formel) 

% aussagenlogische Verbindungen:
:- op(500,yfx,&), op(600,yfx,'\\/'),
    op(600,xfx,'=>'), op(600,xfx,'<=>').

eval(Env,(F & G)) :- !, eval(Env,F) , eval(Env,G).
eval(Env,(F \/ G)) :- !, (eval(Env,F), ! ; eval(Env,G)). 
eval(Env,(F => G)) :-  !, eval(Env,(neg(F) \/ G)).
eval(Env,(F <=> G)) :- !, eval(Env,(F => G) & (G => F)).
eval(Env,neg(F)) :- !, (eval(Env,F) -> fail ; true).

eval(_Env,true) :- !.

% Existenzquantor = Suchen eines passenden Objekts:
eval(Env,ex(Var:Ty,Restr,Fml)) :-
    var(Var), !, % getypt
    datenbank:objekt(Obj:Ty), % suche in der Datenbank;
    eval([(Var,Obj)|Env], Restr & Fml), !. % erste Lösung

eval(Env,all(Var:Ty,Restr,Fml)) :-
    var(Var), !,
    eval(Env, neg(ex(Var:Ty,Restr,neg(Fml)))).

eval(Env,iota(Var:Ty,Restr,Fml)) :- 
    var(Var), !,
    normalize(lam(Var,Restr)*X,RestrX),
    eval(Env,ex(Var:Ty,Restr,
                Fml & all(X:Ty,RestrX,eq(X,Var)))).

eval(Env,most(X:Ty,Restr,Formel)) :-
    var(X), !,
    eval(Env,card(X:Ty,Restr,Formel),NPos),
    eval(Env,card(X:Ty,Restr,neg(Formel)),NNeg),
    NPos > NNeg.

eval(Env,many(X:Ty,Restr,Formel)) :-
    var(X), !,
    eval(Env,card(X:Ty,Restr,Formel),NPos),
    eval(Env,card(X:Ty,Restr,true),NAll),
    N3 is (NPos * 3),
    N3 > NAll.

% Verallgemeinerte Paar-Quantoren:
% ex, all, iota, most, many für Individuenpaare

eval(Env,ex((X,Y):(TyX*TyY),Restr,Fml)) :-
    var(X), var(Y), !, % getypt
    datenbank:objekt(ObjX:TyX), % suche in der Datenbank;
    datenbank:objekt(ObjY:TyY), % nur erste Lösung:
    eval([(X,ObjX),(Y,ObjY)|Env], Restr & Fml), !. 

eval(Env,most((X,Y):(TyX * TyY),Restr,Formel)) :-
    eval(Env,card((X,Y):(TyX*TyY),Restr,Formel),NPos),
    eval(Env,card((X,Y):(TyX*TyY),Restr,neg(Formel)),NNeg),
    NPos > NNeg.

% To do: all, many, iota
eval(Env,(X < Y))  :-
    !, evals(Env,[X,Y],[VX,VY]), VX < VY.     
eval(Env,(X =< Y))  :-
    !, evals(Env,[X,Y],[VX,VY]), VX =< VY.     
eval(Env,eq(X,Y))  :-
    !, evals(Env,[X,Y],[VX,VY]), VX == VY.     

% Anzahlaussagen
eval(Env,anzahl(Var:Ty,Formel,Zahl)) :-
    !, antworten(Env,qu(Var:Ty,true,Formel),Objekte), 
    length(Objekte,Zahl). % N >= Zahl ?

% Grundformeln (der Datenbank)
eval(Env,Formel) :-
    functor(Formel,Pn,Arity),
    Arity > 0, !, 
    Formel =.. [Pn|Args],
    evals(Env,Args,Vals),
    EFormel =.. [Pn|Vals], call(EFormel).

% eval(+Belegung,+Frage,-Werte)
eval(Env,qu(V:Ty,Restr,Formel),Objekte) :-
    var(V),!,
    findall(O,(objekt(O:Ty),
               eval([(V,O)|Env],(Restr & Formel))),Os),
    sort(Os,Objekte).

eval(Env,qu((X,Y):(TyX*TyY),Restr,Formel),Paare) :-
    !, findall((A,B),(objekt(A:TyX),objekt(B:TyY),
         eval([(X,A),(Y,B)|Env],(Restr & Formel))),ABs),
    sort(ABs,Paare).

eval(Env,Formel,Wert) :-
    eval(Env,Formel) -> Wert = ja ; Wert = nein.

