:- module(lambdaDrs,[reduceTerm/2,mergeTerm/2]).
:- op(500,yfx,&), op(600,yfx,'\\/'), op(600,xfx,'=>').

normalize(Term,MergedTerm):-
    reduceTerm(Term,ReducedTerm),
    mergeTerm(ReducedTerm,MergedTerm).

reduceTerm(T,Nf) :-
    beta(T,ConT,Tag), 
    ( Tag = chd -> reduceTerm(ConT,Nf) ; Nf = ConT ).        

reduceTerms([T|Ts],[Nf|Nfs]) :-
    reduceTerm(T,Nf), reduceTerms(Ts,Nfs).
reduceTerms([],[]).

% mergeTerm(+Term,-mergedTerm)

mergeTerm(T,Nf) :-
    mergeDrs(T,MT,Tag),    
    ( Tag = chd -> mergeTerm(MT,Nf) ; Nf = MT ).

mergeTerms([T|Ts],[Nf|Nfs],Tag) :-
    mergeDrs(T,MT,TagT),
    ( TagT = chd -> mergeTerm(MT,Nf) ; Nf = MT ),
    mergeTerms(Ts,Nfs,TagTs),
    ( TagT = not -> Tag = TagTs ; Tag = chd ).    
mergeTerms([],[],not).

% mergeDrs(+ReducedTerm, -mergedTerm, -geaendert?)

mergeDrs(X, X, not) :-  var(X), !.

mergeDrs(T*S, Mrg, Tag) :-
    !, mergeDrs(T,ConT,TagT),
    ( TagT = chd -> Mrg = ConT*S, Tag = chd
    ; mergeDrs(S,ConS,Tag), Mrg = T*ConS ).

mergeDrs(lam(X,D),lam(X,MD),Tag):-
    !, mergeDrs(D,MD,Tag).

mergeDrs(drs(D,F),drs(D,MF),Tag):-
    !, mergeTerms(F,MF,Tag). 

mergeDrs(neg(D),neg(MD),Tag):-
    !, mergeDrs(D,MD,Tag).

mergeDrs((D1 => D2),Mrg,Tag):-
    !, mergeDrs(D1,MD1,Tag1),
    ( Tag1 = chd -> Mrg = (MD1 => D2), Tag = chd
    ; mergeDrs(D2,MD2,Tag),Mrg = (D1 => MD2)).
        
mergeDrs((D1 \/ D2),Mrg,Tag):- 
    !, mergeDrs(D1,MD1,Tag1),
    ( Tag1 = chd -> Mrg = (MD1 \/ D2), Tag = chd
    ; mergeDrs(D2,MD2,Tag),Mrg = (D1 \/ MD2)).
        
mergeDrs(Drs,Mrg,Tag):- 
    Drs =.. [Q,X,D1,D2],
    member(Q,[all,most,many,iota,qu,anzahl]), % (Nicht: ex !)
    var(X),
    !, mergeDrs(D1,MD1,Tag1),
    ( Tag1 = chd -> Mrg =.. [Q,X,MD1,D2], Tag = chd
    ; mergeDrs(D2,MD2,Tag),Mrg =.. [Q,X,MD1,MD2] ).

mergeDrs(Drs1 + Drs2, Mrg, Chd) :-
    var(Drs1), !,
    mergeDrs(Drs2,D2,Chd), Mrg=(Drs1+D2).

mergeDrs(Drs1 + Drs2, Mrg, Chd) :-
    var(Drs2), !,
    mergeDrs(Drs1,D1,Chd), Mrg=(D1+Drs2).

mergeDrs(drs(Rs1,Fs1) + (drs([],Fs2) + D3),Mrg,Chd) :-
    !, mergeDrs(drs(Rs1,Fs1) + drs([],Fs2),Mrg1,Chd1),
    ( Chd1=chd
    -> mergeDrs(Mrg1+D3,Mrg,Chd)
    ; mergeDrs(drs([],Fs2) + D3,Mrg2,Chd2),
        ( Chd2=chd
        -> mergeDrs(drs(Rs1,Fs1) + Mrg2,Mrg,Chd)
        ; Mrg = (drs(Rs1,Fs1) + (drs([],Fs2) + D3)),
            Chd=not
        )
    ).
                
mergeDrs((drs(R1,F1) + drs(R2,F2)),_,_) :-
    not(disjoint(R1,R2)),
    !, write(user,'\nFehler: nicht-disjunkte Refs'),
    write(user,' in '),writen(drs(R1,F1)+drs(R2,F2)),
    fail. % ambiger Skopus von X in top(D1),top(D2)!

mergeDrs((drs([R|R1],F1) + drs(R2,F2)),Mrg,Chd) :-
    !, % disjunkte Refs
    mergeDrs((drs(R1,F1)+drs(R2,F2)),drs(Rs,Fs),Chd),
    Mrg = drs([R|Rs],Fs).

mergeDrs((drs([],F1) + drs(R2,F2)),Mrg,chd) :-
    !, append(F1,F2,F),Mrg = drs(R2,F).

mergeDrs((D1 + D2), Mrg, Tag):-
    !, mergeDrs(D1,MD1,Tag1),
    ( Tag1 = chd -> Mrg = (MD1 + D2), Tag = chd
    ; mergeDrs(D2,MD2,Tag), Mrg = (D1 + MD2) ).

mergeDrs(F,F,not):- !. % einfache Bedingung
        
occurs(X,[V|Vs]) :-
   (X == V -> true ; occurs(X,Vs)).

