% :- module(lambda,[normalize/2]).

% Term := Var | Atom | (Term * Term) | lam(X,Term)
%       | Atom(Term,...,Term) (logischer Term/Formel!)

normalize(T,Nf) :-                
    beta(T,ConT,Tag),     % Einen Reduktionsschritt
    ( Tag = chd           % machen; falls er zu ConT=/=T
    -> normalize(ConT,Nf) % führt, weiter vereinfachen,
    ; Nf = ConT ).        % sonst ist T die Normalform
                          % ConT für lam unter Quantor
normalize_seq([T|Ts],[Nf|Nfs]) :-
    normalize(T,Nf),
    normalize_seq(Ts,Nfs).
normalize_seq([],[]).

% Beta-Reduktion beta(+Term,-Reduziert,-geändert?)

beta(X, X, not) :-  var(X), !.

beta(T*S, T*ConS, Tag) :-
    var(T),
    !, beta(S,ConS,Tag).

beta(lam(X,R)*S, RDup, chd) :- % beta-Regel
    (var(X), !
    ; X = (X1,X2), var(X1), var(X2)), % Plural
    !, alpha(lam(X,R),lam(XDup,RDup)),
    XDup = S.         % simuliert R[X/S]

beta(T*S, U, Tag) :-
    !, beta(T,ConT,TagT),
    ( TagT = chd -> U = ConT*S, Tag = chd
    ; beta(S,ConS,Tag), U = T*ConS ).

beta(lam(X,R), lam(X,ConR), Tag) :-
    !, beta(R,ConR,Tag).

beta(T, ConT, not) :- % für Atom(Term,..,Term) 
    compound(T),      
    !, T =.. [F|Args],
    normalize_seq(Args,Nfs),
    ConT =..[F|Nfs].    

beta(T, T, not) :- !. % Konstante

% Alpha-Reduktion: alpha(+Term,-Umbenannt), 
%                  alpha(+Term,+VarPaare,-Umbenannt)
alpha(T, TDup) :-
    alpha(T, [], TDup), !.

alpha(V, L, VDup) :-
    var(V), !, rename(V,L,VDup). 
alpha(lam(V,R), L, lam(New,RDup)) :-
    var(V), !, alpha(R,[(V,New)|L],RDup).
alpha(Formel,L,FormelDup) :-    
    Formel =.. [Q,V,R,S], % zerlege Formel Q(V,R,S)
    var(V), member(Q,[iota,ex,all,many,most,qu]),
    !, alpha_seq([R,S],[(V,New)|L],[RDup,SDup]),
    % zusammensetzen: Q(New,RDup,SDup)
    FormelDup =.. [Q,New,RDup,SDup].
alpha(T*S, L, TDup*SDup) :-
    !, alpha(T,L,TDup),
    alpha(S,L,SDup).
alpha(C, _, C) :-
    atomic(C), !. 
alpha(Formel,L,FormelDup) :-    
    Formel =.. [Q,(U,V),R,S], % zerlege Formel 
    var(U), var(V), member(Q,[ex,all,many,most,iota,qu]),
    !, alpha_seq([R,S],[(U,UNew),(V,VNew)|L],[RDup,SDup]),
    % zusammensetzen: Q((UNew,VNew),RDup,SDup)
    FormelDup =.. [Q,(UNew,VNew),RDup,SDup].

alpha(T, L, TDup) :-  % für:  Atom(Term,...,Term) 
    compound(T),
    !, T =.. [F|Ts],
    alpha_seq(Ts,L,TsDup),
    TDup =.. [F|TsDup].

alpha_seq([T|Ts], L, [TDup|TsDup]) :-
    alpha(T,L,TDup), alpha_seq(Ts,L,TsDup).
alpha_seq([],_L,[]).

% rename(+Var, +Variablenpaare, -VarUmbenannt)

rename(V, [], V) :- !.
rename(V, [(Var,Dup)|_], Dup) :-
    V == Var, !. 
rename(V, [_|C], Dup) :-  
    !, rename(V,C,Dup).

