:- op(500,yfx,&), op(600,yfx,'\\/'),
    op(600,xfx,'=>'), op(600,xfx,'<=>').

kleinschreibung(Wort,Klein) :-    
    downcase_atom(Wort,Klein).

sem([en(Art,Form), [Vollform]],LamTerm) :-
    wort(Stammform, en(Art,Form), Vollform),
    kleinschreibung(Stammform,LamTerm).

sem([rn(Art,Form), [Vollform]],LamTerm) :-
    wort(Stammform, rn(Art,Form), Vollform),
    kleinschreibung(Stammform,Stamm),
    Formel =.. [Stamm,X,Y], % erzeugt: Stamm(X,Y)
    LamTerm = lam(X,lam(Y,Formel)).

sem([v([nom,akk],Form),[Vollform]],LamTerm) :-
    wort(Stammform,v([nom,akk],Form),Vollform),
    Formel =.. [Stammform,X,Y],
    LamTerm = lam(X,lam(Y,Formel)).

sem([n(Art,Form),[Vollform]],LamTerm) :-
    wort(Stammform, n(Art,Form), Vollform),
    kleinschreibung(Stammform,Stamm),
    Formel =.. [Stamm,X],
    LamTerm = lam(X,Formel).
   
sem([n(Art,Form),RN],LamTerm) :-
    RN = [rn(Art,Form),[_Vollform]],
    !, sem(RN,SemRN),
    normalize(SemRN * X * Y, Formel),
    LamTerm = lam(X,ex(Y,true,Formel)).
   
sem([v([nom],Form),[Vollform]],LamTerm) :-
    wort(Stammform, v([nom],Form), Vollform),
    Formel =.. [Stammform,X],
    LamTerm = lam(X,Formel).
   
sem([det([def], [_,sg,_]),[_Vollform]],
    lam(N,lam(P, iota(X, N*X, P*X)))) :- !. % iota/3
sem([det([indef],[_,sg,_]),[_Vollform]],
    lam(N,lam(P,ex(X, N*X, P*X)))) :- !.    % ex/3
sem([det([all],[_,sg,_]),[_Vollform]],
    lam(N,lam(P,all(X, N*X, P*X)))) :- !.   % all/3

sem([det([def],[_,pl,_]),[Vollform]],
    lam(N,lam(P, anzahl(X,N*X & P*X,Zahl)))) :-
    anzahl(Vollform,Zahl), 2 =< Zahl, !.

sem([det([qu], _Form),[_Welche]],
    lam(N,lam(P,qu(X, N*X, P*X)))).

% für distributive Eigenschaften N und P:
sem([det([def], [_,pl,_]),_Vollform],
    lam(N,lam(P, iota(X, N*X, P*X)))).  % iota/3
sem([det([many],[_,pl,_]),[_Viele]],    % many/3
    lam(N,lam(P,many(X,N*X,P*X)))).
sem([det([most],[_,pl,_]),[_DefArt],[meisten]],
    lam(N,lam(P,most(X,N*X,P*X)))).     % most/3
sem([det([all],[_,pl,_]),[_Vollform]],
    lam(N,lam(P,all(X,N*X,P*X)))).      % all/3

% Quantor vom Typ (1,1,1):
sem([det([more],[-,pl,-]),[mehr]],
    lam(N1,lam(N2,lam(P,more(X,N1*X,N2*X,P*X))))).

sem([a([als(nom)],[komp]),[kleiner]],
    lam(X,lam(Y,(X < Y)))). 
sem([a([als(nom)],[komp]),[größer]],
    lam(X,lam(Y,(Y < X)))). 

% Personalpronomen: freie Variable X!
sem([pron([per],[_Gen,_Num,_Kas]),[_]], lam(P, P*_X)).

sem([pron([rel],_Form),_Vf], lam(P,lam(X,P*X))).

sem([pron([qu],_Form),[Wer]], lam(P,qu(X,true,P*X))) :-
    ( Wer = wer ; Wer = wen ; Wer = was).

sem([poss([def(_,_)],[_Gen,sg,_Kas]),_Poss], 
    lam(RN,lam(P,iota(X, RN*X*_Y, P*X)))).

sem([poss([def(_,_)],[_Gen,pl,_Kas]),_Poss], 
    lam(RN,lam(P,all(X, RN*X*_Y, P*X)))).

sem([poss([qu],[-,-,-]),_Poss],
    lam(RN,lam(P,qu(Y,true,ex(X, RN*X*Y, P*X))))).

sem([poss([rel(_Gen,_Num)],[-,sg,-]),_Poss],   
    lam(RN,lam(P,lam(Y,iota(X, RN*X*Y, P*X))))). 
sem([poss([rel(_Gen,_Num)],[-,pl,-]),_Poss],   
    lam(RN,lam(P,lam(Y,all(X, RN*X*Y, P*X))))).

sem([pron([refl],_),[sich]],
    lam(P,lam(X,P*X*X))) :- !.

sem([pron([rezi],_),[einander]], % Paarabstraktion für
     lam(P,lam((X,Y),P*X*Y & P*Y*X))) :- !. % Symmetrie

sem([np(_Art,_Form),Pron],SemNP) :-
    Pron = [pron(_,_),_Vf], !, sem(Pron,SemNP).

sem([np([def,3,Gen,Num],[Kas]),EN], SemNP) :-
    EN = [en([Gen],[Num,Kas]),[_Name]],
    !, sem(EN,SemEN),
    SemNP = lam(P,P*SemEN).

sem([np(_Art,_Form),EN,SRel], SemNP) :-
    EN = [en(_ArtE,_FormE),[_Name]],
    SRel = [s([rel(_,_)],_) | _Konst],
    !, sem(EN,SemEN), sem(SRel,SemSRel),
    normalize(lam(P,P*SemEN & SemSRel*SemEN),SemNP).

sem([np([def,3,Gen,Num],[Kas]),Det,EN], SemNP) :-
    Det = [det([def],[Gen,Num,Kas]),_],
    EN = [en(_,[Num,nom]),[_Name]],
    !, sem(EN, SemEN),
    SemNP = lam(P, P*SemEN).

sem([np([_,3,_,_],[_]),Det,N], SemNP) :-
    Det = [det(_,_),_], 
    N = [n(_Art,_Form),[Vf]], Vf \= km,
    !, sem(Det,SemDet), sem(N,SemN),
    normalize(SemDet * SemN,SemNP).

sem([np([_,3,_,sg],[Kas]),Det,[n(_,Form),RN]], SemNP) :-
    RN = [rn(_,Form),_Vf], Form = [sg,Kas],
    !, sem(Det,SemDet), sem(RN,SemRN),
    SemN = lam(X,ex(Y,true, SemRN * X * Y)),
    normalize(SemDet * SemN, SemNP).

sem([np([_Def,3,Gen,Num],[_]),Det,N,SRel], SemNP) :-
    Det = [det(_,_)|_], 
    N = [n([Gen],[Num,_]),_N],
    SRel = [s([rel(Gen,Num)],_Form)|_Const], !,
    sem(Det,SemDet), sem(N,SemN), sem(SRel,SemSRel),
    normalize(SemDet * lam(X, SemN*X & SemSRel*X),
              SemNP). % nicht zum Plural (X1,X2) 
                      % spezialisieren?
sem([np(_Art,[_Kas]),Det,RN,NPgen],SemNP) :-
    Det = [det(_,_)|_], RN = [rn(_,_),[_]],
    NPgen = [np(_,[gen])|_],
    !, sem(Det,SemDet), sem(RN,SemRN),
    sem(NPgen,SemNPgen),
    normalize(SemDet *
       lam(X, SemNPgen*lam(Y,(SemRN*X*Y))), SemNP).

sem([np(_Art,[_Kas]),Det,RN,NPgen,RelS],SemNP) :-
    Det = [det(_,_)|_], RN = [rn(_,_),[_]],
    NPgen = [np(_,[gen])|_], 
    RelS = [s([rel(_,_)],_)|_],
    !, sem(Det,SemDet), sem(RN,SemRN), 
    sem(NPgen,SemNPgen), sem(RelS,SemRelS),
    normalize(SemDet *
       lam(X, (SemNPgen*lam(Y,(SemRN*X*Y)))
              & SemRelS*X), SemNP).

sem([np([def,3,_,sg],_),DET,N,EN],SemNP) :-
    DET = [det([def],_),_],
    N = [n(_,_),_], EN = [en(_,_),_],
    !, sem(DET,SemDet), sem(N,SemN), sem(EN,SemEN),
    normalize(SemDet * lam(X,SemN * X & eq(X,SemEN)),
              SemNP).

sem([np([_,3,_,_],[_]),Det,EN,RelS],SemNP) :-
    Det = [det(_,_),_], EN = [en(_,_),_],
    RelS = [s([rel(_,_)],[_,_,_]) | _Konst], !,
    sem(Det,SemDet), sem(EN,SemEN), sem(RelS,SemRelS), 
    normalize(SemDet*lam(X,eq(X,SemEN) & SemRelS*X),
              SemNP).
% TODO: best.Art+N+EN+RelS

sem([np([_Def,3,_,_],[_]),Poss,RN], SemNP) :-
    Poss = [poss(_,_),_], RN = [rn(_,_),_],
    !, sem(Poss,SemPoss), sem(RN,SemRN),
    normalize(SemPoss * SemRN, SemNP).

sem([np([_Def,3,_,_],[_]),Poss,RN,SRel], SemNP) :-
    Poss = [poss(_,_),_], RN = [rn(_,_),_],
    SRel = [s([rel(_,_)],_)|_], 
    !, sem(Poss,SemPoss), sem(RN,SemRN), 
    sem(SRel,SemSRel),
    Sem = lam(P, SemPoss * SemRN
                         * lam(X,P*X & SemSRel*X)),
    normalize(Sem, SemNP).

sem([np([def,3,_,_],_),NP1,[und],NP2], SemNP) :-
   NP1 = [np(_,_)|_], NP2 = [np(_,_)|_], !, 
   sem(NP1,SemNP1), sem(NP2,SemNP2),
   normalize(lam(P, SemNP1 * P & SemNP2 * P),SemNP).

sem([np([def,3,_,_],_),NP1,[oder],NP2], SemNP) :-
   NP1 = [np(_,_)|_], NP2 = [np(_,_)|_], !, 
   sem(NP1,SemNP1), sem(NP2,SemNP2),
   normalize(lam(P, SemNP1 * P \/ SemNP2 * P),SemNP).

sem([np([quant,3,_,pl],_),Det,N1,[als],N2], SemNP) :-
   Det = [det([more],_),_],
   N1 = [n(_,_),_],
   N2 = [n(_,_),_],
   !, sem(Det,SemDet), sem(N1,SemN1), sem(N2,SemN2),
   normalize(SemDet * SemN1 * SemN2, SemNP).

sem([s([_Def],[_,_,vz]), NP1, TV, NP2], SemS)
:-  NP1 = [np([Def1,3,_,_],[Kas1])|_Konst],
    NP2 = [np([Def2,3,_,_],[_])|_],
    TV = [v([nom,akk],[3,_,_,ind]),_Vf],
    sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    ( Kas1 = nom, ((Def2=refl;Def2=rezi) ->
        Sem = SemNP1 * (SemNP2 * SemV) ;                  
        Sem = SemNP1 * lam(X,SemNP2 *
                          lam(Y,SemV * X * Y)))
    ; Kas1 = akk, not(Def1=refl ; Def1=rezi),
        Sem = SemNP1 * lam(Y,SemNP2 *
                          lam(X,SemV * X * Y))
    ), normalize(Sem,SemS).
       
sem([s([_],[Tmp,Mod,vl]), NP1,NP2,TV],
    SemS)
:-  NP1 = [np([_,3,_,_],[Kas1])|_], 
    NP2 = [np([Def2,3,_,_],[_])|_],
    TV = [v([nom,akk],[3,_Num,Tmp,Mod]),_Vf],
    !, sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    ( Kas1 = nom, ((Def2 = refl ; Def2 = rezi)
       -> Sem = SemNP1 * (SemNP2 * SemV)
       ; Sem = 
       SemNP1 * lam(X,SemNP2 * lam(Y,SemV * X * Y)))
    ; Kas1 = akk, Sem =
       SemNP1 * lam(Y,SemNP2 * lam(X,SemV * X * Y))
    ), normalize(Sem,SemS).

sem([s([_Def],[_,_,ve]), TV,NP1,NP2], SemS)
:-  NP1 = [np([Def1,3,_,_],[Kas1])|_Konst], 
    NP2 = [np([Def2,3,_,_],[_])|_],
    TV = [v([nom,akk],[3,_,_,ind]),_Vf],
    !, sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    (Kas1 = nom, ((Def2=refl;Def2=rezi) ->
        Sem = SemNP1 * (SemNP2 * SemV) ;
        Sem = SemNP1 * lam(X,SemNP2 *
                          lam(Y,SemV * X * Y)))
    ; Kas1 = akk, ((Def1=refl ; Def1=rezi) ->
        Sem = SemNP2 * (SemNP1 * SemV) ;                      
        Sem = SemNP1 * lam(Y,SemNP2 *
                          lam(X,SemV * X * Y)))
    ), normalize(Sem,SemS).
       
sem([s([_DefS],_),NP,Kopula,NPpred],LamTerm) :-
    NP = [np(_,_)|_], Kopula = [v([sein],_),[_]],
    NPpred = [np([Def,3,_,_],[nom])|_],
    (Def=indef; Def=def),
    !, sem(NP,SemNP), sem(NPpred,SemPred),
    Pred = lam(X,SemPred*lam(Y,eq(X,Y))),
    normalize(SemNP*Pred, LamTerm).

sem([s([DefS],Form),Kopula,NP,NPpred],LamTerm) :-
   Kopula = [v([sein],_),[_]],
   !, sem([s([DefS],Form),NP,Kopula,NPpred],LamTerm).

sem([s([DefS],Form),NP,NPpred,Kopula],LamTerm) :-
   Kopula = [v([sein],_),[_]],
   !, sem([s([DefS],Form),NP,Kopula,NPpred],LamTerm).

sem([s([_Def],[_,_,vz]), NP1,IV], SemS) :-
    NP1 = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf],
    !, sem(NP1,SemNP1), sem(IV, SemV),
    normalize(SemNP1 * SemV, SemS).

sem([s([_Def],[_,_,ve]), IV,NP1], SemS) :-
    NP1 = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf], 
    !, sem(NP1,SemNP1), sem(IV, SemV),
    normalize(SemNP1 * SemV, SemS).

sem([s([_Def],[_,_,vl]), NP1, IV], SemS) :-
    NP1 = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf], 
    !, sem(NP1,SemNP1), sem(IV, SemV),
    normalize(SemNP1 * SemV, SemS).

% Bedeutung der Nominalphrasen "n km":
sem([n([mask],[pl,_]),[km]],km).

sem([np(_,_),Det,N],SemNP) :-
   Det = [det(_,_),[Vollform]],
   N = [n(_,_),[km]], 
   !, anzahl(Vollform,Zahl), % Umwandlung Atom -> Zahl
   SemNP = lam(P,P*Zahl).

sem([ap(_,_),A,[als],NP],SemAP) :-
   A = [a(_,_),_], NP = [np(_,_)|_], % weiter einschränken?
   !, sem(A,SemA), sem(NP,SemNP),
   Sem = lam(X,SemNP*lam(Y,SemA*X*Y)), 
   normalize(Sem,SemAP).

sem([s(_,_),NP,V,AP], SemS) :-
   NP = [np(_,_)|_], V = [v([sein],_),_], 
   AP = [ap(_,[komp])|_],
   !, sem(NP,SemNP), sem(AP,SemAP),
   normalize(SemNP * SemAP, SemS).

