:- module(drs,[]).

% sem/2 mit lambda-DRS-Termen
:- op(500,yfx,&), op(600,yfx,'\\/'), op(600,xfx,'=>').

kleinschreibung(Wort,Klein) :-    
    downcase_atom(Wort,Klein).

% EN
sem([en(Art,Form), [Vollform]],LamTerm) :-
    !, wort(Stammform, en(Art,Form), Vollform),
    kleinschreibung(Stammform,EN),
    LamTerm = EN.

% VI 
sem([v([nom],Form),[Vollform]],LamTerm) :-
    !, kleinschreibung(Vollform,VollformKl),
    wort(Stammform,v([nom],Form),VollformKl),
    Formel =.. [Stammform,X],
    LamTerm = lam(X,drs([],[Formel])).

% VT 
sem([v([nom,akk],Form),[Vollform]],LamTerm) :-
    !, kleinschreibung(Vollform,VollformKl),
    wort(Stammform,v([nom,akk],Form),VollformKl),
    Formel =.. [Stammform,X,Y],
    LamTerm = lam(X,lam(Y,drs([],[Formel]))).
   
% N
sem([n([Gen],[Num,Kas]),[Vollform]],LamTerm) :-
    !, wort(Stammform, n([Gen],[Num,Kas]), Vollform),
    kleinschreibung(Stammform,Stamm),
    Formel =.. [Stamm,X],
    LamTerm = lam(X,drs([],[ant(X,Gen,Num),Formel])).
   
% RN 
sem([rn([Gen],[Num,Kas]), [Vollform]],LamTerm) :-
    !, wort(Stammform, rn([Gen],[Num,Kas]), Vollform),
    kleinschreibung(Stammform,Stamm),
    Formel =.. [Stamm,X,Y],
    LamTerm = lam(X,lam(Y,
                  drs([],[ant(X,Gen,Num),Formel]))).

% N aus RN
sem([n(Art,Form),[rn(Art,Form),[Vollform]]],LamTerm) :-
    !, sem([rn(Art,Form),[Vollform]],SemRN),
    normalize(SemRN * X * Y, Drs),
    LamTerm = lam(X,drs([Y],[])+Drs).
   
% DET. Bei all: distributive sg-Lesarten 
sem([det([all],[_,sg,_]),_Vollform],
    lam(N,lam(P,drs([],[all(X,N*X,P*X)])))) :- !. 

sem([det([many],[_,pl,_]),_Vollform],
    lam(N,lam(P,drs([],[many(X,N*X,P*X)])))) :- !. 
sem([det([most],[_,pl,_]),_Vollform],
    lam(N,lam(P,drs([],[most(X,N*X,P*X)])))) :- !. 

% DET = the 
sem([det([def],[_,sg,_]),_Vollform],
    lam(N,lam(P, drs([],[iota(X,N*X,P*X)])))) :- !.

% DET = k (Anzahldeterminator) (TODO: Refs?)
sem([det([def],[_,pl,_]),[Vollform]],
    lam(N,lam(P, drs([],[anzahl(X,N*X+P*X,Zahl)])))) :-
    anzahl(Vollform,Zahl), !.

% DET = a
sem([det([indef],[_,sg,_]),_Vollform], 
    lam(N,lam(P, drs([X],[]) + (N*X + P*X)))) :- !.
% DET = einige (3 Lesarten)
sem([det([indef],[_,pl,_]),_Vollform], 
    lam(N,lam(P, drs([X],[]) + (N*X + P*X)))).
/*
sem([det([indef],[_,pl,_]),_Vollform],
    lam(N,lam(P, drs([X,Y],[]) + (N*X + N*Y + P*X*Y)))). 

sem([det([indef],[_,pl,_]),_Vollform],
    lam(N,lam(P, drs([(X,Y)],[]) + (N*X + N*Y + P*X + P*Y)))). 
sem([det([indef],[_,pl,_]),_Vollform],
    lam(N,lam(P, drs([(X,Y)],[]) + (N*X + N*Y + P*X*Y)))). % 3.Lesart folgt
*/
% DET = die.pl, distributiv individuelle Lesart
sem([det([def], [_,pl,_]),_],
    lam(N,lam(P,iota(X,N*X,P*X)))). 

% TODO: interrogative drs?, sg/pl-Lesarten?
sem([det([qu], _Form),[Vollform]],
    lam(N,lam(P,qu(X,N*X,P*X)))) :-
    concat_atom(['welche'|_],Vollform), !.

% PRON [per]
sem([pron([per],[Gen,Num,_Kas]),_Pron],
    lam(P, drs([X],[anp(X,Gen,Num)]) + P*X)) :- !.

sem([pron([rel],_Form),_Rel], lam(P,lam(X,P*X))) :- !.
% TODO: interrogative DRS?
sem([pron([qu],_Form),_Pron], lam(P,qu(X,P*X))) :- !.

% POSS (ohne Eindeutigkeitsbedingung) 
sem([poss([def(Gen,Num)],_Form),_Vollform],Sem):-
    !, Sem = lam(RN,lam(P, drs([X,Y],
                               [anposs(Y,Gen,Num)])
                           + (RN*X*Y + P*X))).
sem([pron([refl],_Form),[sich]],
    lam(P,lam(X, P*X*X))) :- !.

sem([pron([rezi],_Form),[einander]], 
    lam(P,lam((X,Y), P*X*Y + P*Y*X))) :- !.

% NP = PRON (personal, reflexiv, reziprok, interrogativ)
sem([np(_Art,_Form),Pron],SemNP) :-
    Pron = [pron([_ArtP],_FormP),_],
    !, sem(Pron,SemNP).

% NP = EN (mit neuem Diskursreferenten)
sem([np([def,3,Gen,Num],[Kas]),EN], SemNP) :-
    EN = [en([Gen],[Num,Kas]),_Vf],
    !, sem(EN,SemEN),
    SemNP = lam(P,drs([X],[ant(X,Gen,Num),eq(X,SemEN)])
                  + P*X).
% NP = DET EN 
sem([np([def,3,Gen,Num],[_Kas]),Det,EN], SemNP) :-
    Det = [det([def],_),_],
    EN = [en([Gen],[Num,nom]),_Vf],        
    !, sem([np([def,3,Gen,Num],[nom]),EN], SemNP).

% NP = DET+N+PN 
sem([np([def,3,_Gen,_Num],[_Kas]),
     [det([def],_),_],N,EN], SemNP) :-
    N = [n(_,_),_], EN = [en(_,_),_],
    !, sem(N,SemN), sem(EN,SemEN),
    normalize(lam(P, drs([X],[eq(X,SemEN)])
                     + (SemN*X + P*X)), SemNP).

% NP = DET+N 
sem([np([_,3,_,_],[_]),Det,N], SemNP) :-
    Det = [det(_,_)|_],
    N = [n(_,_),[Vf]], Vf \= km,
    !, sem(Det,SemDet), sem(N,SemN),
    normalize(SemDet * SemN,SemNP).

% NP = DET N RelS
sem([np([_,3,_,_],[_]),Det,N,RelS], Sem) :-
    N = [n(_,_),_], RelS = [s([rel(_,_)],_)|_],
    !, sem(Det,SemDet),
    sem(N,SemN), sem(RelS,SemRelS), 
    SemNS = lam(X, (SemN*X + SemRelS*X)),
    normalize(SemDet*SemNS,Sem).

sem([np([_,3,_,_],[_]),
       Det,[rn(Art,Form),RN],NPgen,RelS], Sem) :-
    RelS = [s([rel(_,_)],_)|_],
    NPgen = [np(_,[gen])|_],
    !, sem(Det,SemDet),   
    sem([rn(Art,Form),RN],SemRN),
    sem(NPgen,SemNPgen),
    sem(RelS,SemRelS), 
    SemN = lam(X, SemNPgen*lam(Y,(SemRN*X)*Y + SemRelS*X)),
    normalize(SemDet*SemN,Sem).

% NP = POSS+RN  
sem([np([def,3,Gen,Num],[Kas]), Poss, RN], SemNP) :-
    Poss = [poss([def(_,_)],[Gen,Num,Kas]),_],
    RN = [rn([Gen],[Num,Kas]),_],
    !, sem(Poss, SemPoss), sem(RN, SemRN),
    normalize(SemPoss * SemRN, SemNP).

% NP = Zahl km (TODO: testen)
sem([np([_,3,_,_],[_]),
       [det([def],_),[Zahlatom]],[n(_,_),[km]]],
    lam(P,drs([],[])+P*Zahl)) :-  
    anzahl(Zahlatom,Zahl), !. % Ausnahme: Maß

% NP = DET RN NPgen
sem([np([_,3,_,_],[_]),Det,RN,NPgen],Sem) :-
    Det = [det(_,_)|_],
    RN = [rn(_,_),_],
    NPgen = [np([_,3,_,_],[gen])|_],
    !, sem(Det,SemDet),
    sem(RN,SemRN), sem(NPgen,SemNPgen),
    SemN = lam(X,SemNPgen*lam(Y,(SemRN*X)*Y)),
    normalize(SemDet*SemN,Sem).

sem([np([def,3,_,_],_),NP1,[und],NP2], SemNP) :-
   NP1 = [np(_,_)|_], NP2 = [np(_,_)|_],
   !, sem(NP1,SemNP1), sem(NP2,SemNP2),
   normalize(lam(P, (SemNP1 * P) + (SemNP2 * P)),SemNP).

sem([np([def,3,_,_],_),NP1,[oder],NP2], SemNP) :-
   NP1 = [np(_,_)|_], NP2 = [np(_,_)|_],
   !, sem(NP1,SemNP1), sem(NP2,SemNP2),
   normalize(lam(P, (SemNP1 * P) \/ (SemNP2 * P)),SemNP).

% S = NP+VI 
sem([s([_Def],[_,_,vz]), NP,IV], SemS) :-
    NP = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf],
    !, sem(NP,SemNP), sem(IV, SemV),
    normalize(SemNP * SemV, SemS).

sem([s([_Def],[_,_,vl]), NP, IV], SemS) :-
    NP = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf], 
    !, sem(NP,SemNP), sem(IV, SemV),
    normalize(SemNP * SemV, SemS).

sem([s([_Def],[_,_,ve]), IV,NP], SemS) :-
    NP = [np([_,3,_,Num],[nom])|_Konst],
    IV = [v([nom],[3,Num,_,ind]),_Vf], 
    !, sem(NP,SemNP), sem(IV, SemV),
    normalize(SemNP * SemV, SemS).

sem([s([_Def],[_,_,ve]),TV,NP1,NP2], SemS)
:-  TV = [v([nom,akk],[3,_Num,_,ind]),_V],
    NP1 = [np([Def1,3,_,_],[Kas1])|_], 
    NP2 = [np([Def2,3,_,_],[_])|_],
    !, sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    (Kas1 = nom, ( (Def2 = rezi ; Def2 = refl) 
        -> Sem = SemNP1 * (SemNP2 * SemV)
        ; Sem = SemNP1 * lam(X,SemNP2 *
                            lam(Y,SemV * X * Y)))
    ; Kas1 = akk, ( (Def1 = rezi ; Def1 = refl) 
        -> Sem = SemNP2 * (SemNP1 * SemV)
        ; Sem = SemNP1 * lam(Y,SemNP2 *
                          lam(X,SemV * X * Y)) )
    ), normalize(Sem,SemS).
sem([s([_Def],[_,_,vz]),NP1,TV,NP2], SemS)
:-  NP1 = [np([Def1,3,_,_],[Kas1])|_], 
    TV =  [v([nom,akk],[3,_Num,_,ind]),_V],
    NP2 = [np([Def2,3,_,_],[_])|_],
    !, sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    ( Kas1 = nom, ( (Def2 = rezi ; Def2 = refl) 
        -> Sem = SemNP1 * (SemNP2 * SemV)
        ; Sem = SemNP1 * lam(X,SemNP2 *
                            lam(Y,SemV * X * Y)))
    ; Kas1 = akk, ( (Def1 = rezi ; Def1 = refl)
        -> Sem = SemNP2 * (SemNP1 * SemV)
        ; Sem = SemNP1 * lam(Y,SemNP2 *
                          lam(X,SemV * X * Y)) )
    ), normalize(Sem,SemS).

sem([s([_],[_,_,vl]),NP1,NP2,TV], SemS)
:-  NP1 = [np([_,3,_,_],[Kas1])|_], 
    NP2 = [np([Def2,3,_,_],[_])|_],
    TV = [v([nom,akk],[3,_Num,_,_]),_V],
    !, sem(NP1,SemNP1), sem(NP2,SemNP2), sem(TV,SemV),
    ( Kas1 = nom, ((Def2 = refl ; Def2 = rezi)
       -> Sem = SemNP1 * (SemNP2 * SemV)
       ; Sem = 
       SemNP1 * lam(X,SemNP2 * lam(Y,SemV * X * Y)))
    ; Kas1 = akk, Sem =
      SemNP1 * lam(Y,SemNP2 * lam(X,SemV * X * Y))
    ),
    normalize(Sem,SemS).

% D = S D
sem([d([_Def]),S,D], Sem) :-
    !, sem(S,SemS),
    sem(D,SemD),
    ( debugging(resolve)
    -> resolveDrs:resolve(SemS,SemD,Sem) % später
    ; normalize(SemS + SemD, Sem)
    ).

% D = [] (Diskursende)
sem([d([_Def])], drs([],[])) :- !.

normalize(Term,Reduced):-
    lambdaDrs:normalize(Term,Reduced).


% type(+TypKontext,+Drs,-DrsGetypt,-Typ) :-
type(Kontext,Drs,DrsGetypt,Typ) :-
    typeDrs:type(Kontext,Drs,DrsGetypt,Typ).

% S = D1 \/ D2
sem([s([def],_),S1,[oder],S2], Sem) :-
    S1 = [s([def],_)|_], S2 = [s([def],_)|_],
    !, sem(S1,SemS1), sem(S2,SemS2),
%    normalize(drs([],SemS1 \/ SemS2),Sem).  % TODO
    Sem = drs([],[SemS1 \/ SemS2]). 
% S = D1 + D2
sem([s([def],_),S1,[und],S2], Sem) :-
    S1 = [s([def],_)|_], S2 = [s([def],_)|_],
    !, sem(S1,SemS1), sem(S2,SemS2),
    normalize(SemS1 + SemS2, Sem).  

% S = wenn S1, dann S2
sem([s([def],_),[wenn],S1,[dann],S2], Sem) :-
    S1 = [s(_,_)|_], S2 = [s(_,_)|_], % Defs?
    !, sem(S1,SemS1), sem(S2,SemS2),
    normalize(drs([],[SemS1 => SemS2]), Sem).  % TODO
% S = wenn S1, S2
sem([s([def],_),[wenn],S1,S2], Sem) :-
    S1 = [s(_,_)|_], S2 = [s(_,_)|_], % Defs?
    !, sem(S1,SemS1), sem(S2,SemS2),
    normalize(drs([],[SemS1 => SemS2]), Sem).  % TODO
