% type(+Kontext,+LamTerm,-LamTermGetypt,-Typ)

% Lambda terms:
type([V:Ty|Kontext],X,X,Typ) :-
    var(X),
    !, ( X == V -> Typ = Ty ; type(Kontext,X,X,Typ) ).
type([],X,_,_) :-
    var(X), !, fail.
type(Kontext,(R * S),(RTy * STy),Typ) :-
    !, type(Kontext,S,STy,ArgT),
    type(Kontext,R,RTy,FunT),
    (unify_with_occurs_check(FunT,(ArgT -> Typ))
    -> true  % write errors to output-file?
    ; % format('\nFehler: Argument ~w hat Typ ~w',[S,ArgT]),
      % format('\n        Funktion ~w hat Typ ~w',[R,FunT]),
      fail).
type(Kontext,lam(X,Term),lam(X:Ty,TermTy),Typ) :-
    var(X),
    !, type([X:Ty|Kontext],Term,TermTy,ResultTy),
    Typ = (Ty -> ResultTy).
type(Kontext,lam((X,Y),Term),lam((X,Y):Ty,TermTy),Typ) :-
    !, type([X:TyX,Y:TyY|Kontext],Term,TermTy,ResultTy),
    Ty = (TyX * TyY), (TyX = TyY), % symmetrische Prädikate!
    Typ = (Ty -> ResultTy).
% Boolean combinations of formulas:
type(Kontext,neg(Fml),neg(FmlTy),t) :-
    !, type(Kontext,Fml,FmlTy,t).
type(Kontext,(Fml1 & Fml2),(Fml1Ty & Fml2Ty),t) :-
    !, type(Kontext,Fml1,Fml1Ty,t),
    type(Kontext,Fml2,Fml2Ty,t). 
type(Kontext,(Fml1 \/ Fml2),(Fml1Ty \/ Fml2Ty),t) :-
    !, type(Kontext,Fml1,Fml1Ty,t),
    type(Kontext,Fml2,Fml2Ty,t). 
type(Kontext,(Fml1 => Fml2),(Fml1Ty => Fml2Ty),t) :-
    !, type(Kontext,Fml1,Fml1Ty,t),
    type(Kontext,Fml2,Fml2Ty,t). 
type(Kontext,(Fml1 <=> Fml2),(Fml1Ty <=> Fml2Ty),t) :-
    !, type(Kontext,Fml1,Fml1Ty,t),
    type(Kontext,Fml2,Fml2Ty,t). 

type(_,true,true,t) :- !.  % Bereich: true=unbeschränkt

% Verallgemeinerte Quantoren Q(Var,Restriction,Body) 
type(K,Formel,FormelTy,t) :-    
    Formel =.. [Q,X,R,B], % zerlege Formel Q(X,R,B)
    var(X), member(Q,[iota,ex,all,many,most]),
    !, type(K,lam(X,R & B),lam(X:Ty,RTy & BTy),(Ty -> t)),
    FormelTy =.. [Q,X:Ty,RTy,BTy].

type(K,Formel,FormelTy,t) :-  % mit Paarquantoren   
    Formel =.. [Q,(X,Y),R,B], % zerlege Q((X,Y),R,B)
    var(X), var(Y), 
    member(Q,[iota,ex,all,many,most]),
    !, type(K,lam((X,Y),R & B),
              lam((X,Y):Ty,RTy & BTy),(Ty -> t)),
    FormelTy =.. [Q,(X,Y):Ty,RTy,BTy].

% TODO: Verallg.Quantoren Q(Var,Restr1,Restr2,Body) (more)
type(K,anzahl(X,Fml,N),anzahl(X:Ty,FmlTy,NTy),t) :-
    !, type(K,lam(X,Fml),lam(X:Ty,FmlTy),(Ty -> t)),
    type(K,N,NTy,n).  % Typ n := number

% Questions: unary/binary Wh(Vars,Restriction,Formula)
type(K,qu(X,R,Term),qu(X:Ty,RTy,TermTy),list(Ty)) :- 
    var(X), !, type(K,lam(X,R),lam(X:Ty,RTy),(Ty -> t)),
    type(K,lam(X,Term),lam(X:Ty,TermTy),(Ty -> t)).
type(K,qu((X,Y),R,Tm),qu((X,Y):Ty,RTy,TmTy),list(Ty)) :- 
    !, type(K,lam((X,Y),R),lam((X,Y):Ty,RTy),(Ty -> t)),
    type(K,lam((X,Y),Tm),lam((X,Y):Ty,TmTy),(Ty -> t)).

% Algebraic and atomic terms:
type(_Kontext,Modul:Term,Modul:Term,Typ) :- 
    !, Typ = t.         % Grundaussagen der Datenbank
type(_Kontext,N,N,Typ) :-
    number(N), !, Typ = n.
type(Kontext,(X,Y),(XTy,YTy),Typ) :- % Variablentupel 
    !, type(Kontext,X,XTy,TyX),
    type(Kontext,Y,YTy,TyY),
    Typ = (TyX * TyY).
type(Kontext,Term,Term,Typ) :- 
    Term =.. [F|Args],
    !, types(Kontext,Args,[],Typen),
    ( Typen = [],         TypF = Typ
    ; Typen = [T],        TypF = (T -> Typ)
    ; Typen = [T1,T2],    TypF = (T1*T2 -> Typ) ),
    functor(Term,F,Arity),
    (setof(Ty,datenbank:type(F/Arity,Ty),TypDecs)
    -> member(TypF,TypDecs) /* keine Fehlermeldung */
    ; format('\ndatenbank:~w hat keinen Typ\n',[F/Arity]),
        fail
    ).

% Typing a list of arguments
types(Kontext,[A|Args],Acc,TypArgs) :-
    type(Kontext,A,_,TypA),
    types(Kontext,Args,[TypA|Acc],TypArgs).
types(_Kontext,[],Typen,TypArgs) :-
    reverse(Typen,TypArgs).

% Erzeuge Typkontext für die freien Variablen
type(Term,Typed,Type) :-
    free(Term,Vars),
    mkContext(Vars,Kontext),
    type(Kontext,Term,Typed,Type).

mkContext([],[]).
mkContext([X|Xs],[X:_TyX|XsTys]) :-  % Typvariable _TyX
    mkContext(Xs,XsTys).

free(Term,Free) :-
    free(Term,[],Free).

free(T,Acc,Acc) :- atomic(T), !.
free(X,Acc,AccX) :-
    var(X), !, sort([X|Acc],AccX). % keine Duplikate
free((S*T),Acc,Xs) :-
    !, free(S,Acc,AccS),
    free(T,AccS,AccT), sort(AccT,Xs). 
free(lam(X,T),Acc,Free) :-
    !, free(T,[],FreeT),
    ( var(X) -> 
        (setof(Y,(member(Y,FreeT),not(Y==X)),Vs)
        -> FreeLam = Vs ; FreeLam = [])
    ; X = (X1,X2), var(X1), var(X2),
        (setof(Y,(member(Y,FreeT),
                  not(Y==X1),not(Y==X2)),Vs)
        -> FreeLam = Vs ; FreeLam = [])
    ), append(FreeLam,Acc,FreeLamAcc),
    sort(FreeLamAcc,Free).
free(FmlA & FmlB, Acc, AccFml) :-
    !, free(FmlA,Acc,AccA),
    free(FmlB,AccA,AccAB),
    sort(AccAB,AccFml).
free(FmlA => FmlB, Acc, AccFml) :-
    !, free(FmlA & FmlB, Acc, AccFml).
free(FmlA \/ FmlB, Acc, AccFml) :-
    !, free(FmlA & FmlB, Acc, AccFml).
free(not(Fml),Acc,Free) :-
    !, free(Fml,Acc,Free).
free(Binder,Acc,Free) :-  
    Binder =.. [F,Vars,Rest,Fml],
    member(F,[ex,all,most,many,iota,qu]),  % add: more
    !, free(lam(Vars,Rest&Fml),Acc,Free).
free(anzahl(Vars,Fml,_Number),Acc,Free) :-  
    !, free(lam(Vars,Fml),Acc,Free).
free(Fml,Acc,Free) :-   % atomic Fml or algebr.Term
    Fml =.. [_F|Args],
    !, frees(Args,Acc,Free).

frees([Trm|Trms],Acc,Free) :-
    !, free(Trm,Acc,FreeTrm),
    frees(Trms,FreeTrm,Free).
frees([],Acc,Free) :-
    !, sort(Acc,Free).

